@@ -21,6 +21,9 @@
 */
 
 #include <config.h>
+#include <glib.h>
+#include <libgnome/gnome-defs.h>
+#include <libgnome/gnome-i18n.h>
 #include <gal/widgets/e-scroll-frame.h>
 
 #include "gtkhtml.h"
@@ -29,16 +32,18 @@
 #include "htmlclueflow.h"
 #include "htmlcursor.h"
 #include "htmlengine.h"
+#include "htmlengine-edit.h"
 #include "htmlengine-edit-cursor.h"
 #include "htmlengine-edit-movement.h"
 #include "htmlengine-edit-selection-updater.h"
 #include "htmlobject.h"
 #include "htmlselection.h"
 
+#include "menubar.h"
 #include "spell.h"
 
-#define DICTIONARY_IID "OAFIID:GNOME_Spell_Dictionary:0.1"
-#define CONTROL_IID "OAFIID:GNOME_Spell_Control:0.1"
+#define DICTIONARY_IID "OAFIID:GNOME_Spell_Dictionary:0.2"
+#define CONTROL_IID "OAFIID:GNOME_Spell_Control:0.2"
 
 struct _SpellPopup {
 	GtkHTMLControlData *cd;
@@ -68,7 +73,7 @@ destroy (GtkWidget *w, SpellPopup *sp)
 
 			gtk_clist_get_text (GTK_CLIST (sp->clist),
 					    GPOINTER_TO_INT (GTK_CLIST (sp->clist)->selection->data), 0, &replacement);
-			html_engine_replace_word_with (sp->cd->html->engine, replacement);
+			html_engine_replace_spell_word_with (sp->cd->html->engine, replacement);
 
 			/* printf ("replace: %s with: %s\n", sp->misspeled_word, replacement); */
 		}
@@ -119,7 +124,7 @@ fill_suggestion_clist (GtkWidget *clist, const gchar *word, GtkHTMLControlData *
 	CORBA_exception_init (&ev);
 	seq = GNOME_Spell_Dictionary_getSuggestions (cd->dict, word, &ev );
 
-	if (ev._major == CORBA_NO_EXCEPTION) {
+	if (ev._major != CORBA_SYSTEM_EXCEPTION) {
 		for (i=0; i < seq->_length; i++) {
 			suggested_word [0] = seq->_buffer [i];
 			gtk_clist_append (GTK_CLIST (clist), (gchar **) suggested_word);
@@ -205,9 +210,10 @@ spell_check_word (GtkHTML *html, const gchar *word, gpointer data)
 	if (!cd->dict)
 		return TRUE;
 
+	/* printf ("check word: %s\n", word); */
 	CORBA_exception_init (&ev);
 	rv = GNOME_Spell_Dictionary_checkWord (cd->dict, word, &ev);
-	if (ev._major != CORBA_NO_EXCEPTION)
+	if (ev._major == CORBA_SYSTEM_EXCEPTION)
 		rv = TRUE;
 	CORBA_exception_free (&ev);
 
@@ -255,23 +261,18 @@ spell_set_language (GtkHTML *html, const gchar *language, gpointer data)
 	if (!cd->dict)
 		return;
 
+	/* printf ("spell_set_language %s\n", language); */
+
 	CORBA_exception_init (&ev);
-	GNOME_Spell_Dictionary_setTag (cd->dict, "language-tag", language, &ev);
+	GNOME_Spell_Dictionary_setLanguage (cd->dict, language, &ev);
 	CORBA_exception_free (&ev);
+
+	menubar_set_languages (cd, language);
 }
 
 void
 spell_init (GtkHTML *html, GtkHTMLControlData *cd)
 {
-	CORBA_Environment   ev;
-
-	if (!cd->dict)
-		return;
-
-	CORBA_exception_init (&ev);
-	/* GNOME_Spell_Dictionary_setTag (cd->dict, "sug-mode", "fast", &ev); */
-	GNOME_Spell_Dictionary_setTag (cd->dict, "encoding", "utf-8", &ev);
-	CORBA_exception_free (&ev);
 }
 
 static void
@@ -280,9 +281,9 @@ set_word (GtkHTMLControlData *cd)
 	CORBA_Environment ev;
 
 	CORBA_exception_init (&ev);
-	gtk_html_select_word (cd->html);
+	html_engine_select_spell_word_editable (cd->html->engine);
 	bonobo_property_bag_client_set_value_string (cd->spell_control_pb, "word",
-						     html_engine_get_word (cd->html->engine), &ev);
+						     html_engine_get_spell_word (cd->html->engine), &ev);
 	CORBA_exception_free (&ev);
 }
 
@@ -290,7 +291,8 @@ static gboolean
 next_word (GtkHTMLControlData *cd)
 {
 	gboolean rv = TRUE;
-	while (html_engine_forward_word (cd->html->engine) && (rv = html_engine_word_is_valid (cd->html->engine)))
+	while (html_engine_forward_word (cd->html->engine)
+	       && (rv = html_engine_spell_word_is_valid (cd->html->engine)))
 		;
 
 	return rv;
@@ -323,7 +325,7 @@ replace_cb (BonoboListener    *listener,
 
 	/* printf ("replace '%s'\n", BONOBO_ARG_GET_STRING (arg)); */
 
-	html_engine_replace_word_with (cd->html->engine, BONOBO_ARG_GET_STRING (arg));
+	html_engine_replace_spell_word_with (cd->html->engine, BONOBO_ARG_GET_STRING (arg));
 	check_next_word (cd, FALSE);
 }
 
@@ -347,7 +349,7 @@ add_cb (BonoboListener    *listener,
 	GtkHTMLControlData *cd = (GtkHTMLControlData *) user_data;
 	gchar *word;
 
-	word = html_engine_get_word (cd->html->engine);
+	word = html_engine_get_spell_word (cd->html->engine);
 	g_return_if_fail (word);
 
 	GNOME_Spell_Dictionary_addWordToPersonal (cd->dict, word, ev);
@@ -365,7 +367,7 @@ ignore_cb (BonoboListener    *listener,
 	GtkHTMLControlData *cd = (GtkHTMLControlData *) user_data;
 	gchar *word;
 
-	word = html_engine_get_word (cd->html->engine);
+	word = html_engine_get_spell_word (cd->html->engine);
 	g_return_if_fail (word);
 
 	GNOME_Spell_Dictionary_addWordToSession (cd->dict, word, ev);
@@ -402,7 +404,7 @@ spell_check_dialog (GtkHTMLControlData *cd, gboolean whole_document)
 		html_engine_beginning_of_document (cd->html->engine);
 	}
 
-	if (html_engine_word_is_valid (cd->html->engine))
+	if (html_engine_spell_word_is_valid (cd->html->engine))
 		if (next_word (cd)) {
 			html_engine_hide_cursor (cd->html->engine);
 			html_cursor_jump_to_position (cd->html->engine->cursor, cd->html->engine, position);
@@ -439,3 +441,78 @@ spell_check_dialog (GtkHTMLControlData *cd, gboolean whole_document)
 	gtk_container_add (GTK_CONTAINER (GNOME_DIALOG (dialog)->vbox), control);
 	gnome_dialog_run_and_close (GNOME_DIALOG (dialog));
 }
+
+static void
+language_cb (BonoboUIComponent *uic, const char *path, Bonobo_UIComponent_EventType type,
+	     const char *state, gpointer user_data)
+{
+	GtkHTMLControlData *cd = (GtkHTMLControlData *) user_data;
+	GString *str, *lang;
+	gchar *val;
+	gint i;
+
+	if (cd->block_language_changes || !cd->languages)
+		return;
+
+	/* printf ("language callback %s\n", path); */
+
+	str = g_string_new (NULL);
+	lang = g_string_new (NULL);
+	for (i = 0; i < cd->languages->_length; i ++) {
+		g_string_sprintf (lang, "/commands/SpellLanguage%d", i + 1);
+		val = bonobo_ui_component_get_prop (cd->uic, lang->str, "state", NULL);
+		if (val && *val == '1') {
+			g_string_append (str, cd->languages->_buffer [i].abrev);
+			g_string_append_c (str, ' ');
+		}
+	}
+	html_engine_set_language (cd->html->engine, str->str);
+	g_string_free (str, TRUE);
+	g_string_free (lang, TRUE);
+}
+
+void
+spell_create_language_menu (GtkHTMLControlData *cd)
+{
+	CORBA_sequence_GNOME_Spell_Language *seq;
+	CORBA_Environment ev;
+
+	if (cd->dict == CORBA_OBJECT_NIL)
+		return;
+
+	CORBA_exception_init (&ev);
+	cd->languages = seq = GNOME_Spell_Dictionary_getLanguages (cd->dict, &ev);
+
+	if (ev._major == CORBA_NO_EXCEPTION) {
+		if (seq && seq->_length > 0) {
+			GString *str;
+			gchar *line;
+			gint i;
+
+			str = g_string_new ("<submenu name=\"EditSpellLanguagesSubmenu\" _label=\"");
+			g_string_append (str, _("Spell Checking _Languages"));
+			g_string_append (str, "\">\n");
+			for (i = 0; i < seq->_length; i ++) {
+				line = g_strdup_printf ("<menuitem name=\"SpellLanguage%d\" _label=\"%s\""
+							" verb=\"SpellLanguage%d\" type=\"toggle\"/>\n",
+							i + 1, seq->_buffer [i].name, i + 1);
+				g_string_append (str, line);
+				g_free (line);
+			}
+			g_string_append (str, "</submenu>\n");
+
+			bonobo_ui_component_set_translate (cd->uic, "/menu/Edit/EditMisc/EditSpellLanguages/",
+							   str->str, NULL);
+
+			for (i = 0; i < seq->_length; i ++) {
+				g_string_sprintf (str, "SpellLanguage%d", i + 1);
+				bonobo_ui_component_add_listener (cd->uic, str->str, language_cb, cd);
+			}
+			g_string_free (str, TRUE);
+		}
+	} else {
+		g_warning ("CORBA exception: %s\n", bonobo_exception_get_text (&ev));
+		cd->languages = NULL;
+	}
+	CORBA_exception_free (&ev);
+}
